#!/usr/bin/python
# *************************************************************************
#  NORway.py v0.3
#
# Teensy++ 2.0 modifications by judges@eEcho.com
# samsung modifications by foobar
# *************************************************************************
#  noralizer.py - NOR flasher for PS3
#
# Copyright (C) 2010-2011  Hector Martin "marcan" <hector@marcansoft.com>
#
# This code is licensed to you under the terms of the GNU GPL, version 2;
# see file COPYING or http://www.gnu.org/licenses/old-licenses/gpl-2.0.txt
# *************************************************************************

import serial, time, datetime, sys, struct

class TeensySerialError(Exception):
	pass

class TeensySerial(object):
	BUFSIZE = 32768

	def __init__(self, port):
		self.ser = serial.Serial(port, 9600, timeout=5, rtscts=0, dsrdtr=1, xonxoff=0)
		if self.ser is None:
			raise TeensySerialError("could not open serial %s")%port
		self.ser.setDTR()
		self.obuf = ""

	def write(self, s):
		if isinstance(s,int):
			s = chr(s)
		elif isinstance(s,tuple) or isinstance(s,list):
			s = ''.join([chr(c) for c in s])
		self.obuf += s
		while len(self.obuf) > self.BUFSIZE:
			self.ser.write(self.obuf[:self.BUFSIZE])
			self.obuf = self.obuf[self.BUFSIZE:]

	def loadat(self, a, dat):
		self.obuf += chr(CMD_ADDR | (a >> 16))
		self.obuf += chr((a >> 8) & 0xff)
		self.obuf += chr(a & 0xff)
		self.obuf += chr(CMD_WRITE)
		self.obuf += chr((dat >> 8) & 0xff)
		self.obuf += chr(dat & 0xff)

	def flush(self):
		if len(self.obuf):
			self.ser.write(self.obuf)
			self.obuf = ""
		self.ser.flush()

	def read(self, size):
		self.flush()
		data = self.ser.read(size)
		return data

	def readbyte(self):
		return ord(self.read(1))

class NORError(Exception):
	pass

STATUS_DRIVE = 0x40
STATUS_TRIST_N = 0x20
STATUS_RESET_N = 0x10
STATUS_READY = 0x08
STATUS_CE_N = 0x04
STATUS_WE_N = 0x02
STATUS_OE_N = 0x01

# COMMAND LIST!
CMD_NOP = 0x00
CMD_READSTATE = 0x01
CMD_PING1 = 0x02
CMD_PING2 = 0x03
CMD_DRIVE = 0x04
CMD_TRISTATE = 0x06
CMD_RESET = 0x08
CMD_WAIT = 0x0E
CMD_READ = 0x10
CMD_WRITE = 0x18
CMD_READ_ONCE = 0x20
CMD_WRITE_BULK = 0x21
CMD_DELAY = 0x40
CMD_ADDR = 0x80

RES_PONG1 = 0x42
RES_PONG2 = 0xbd

class NORFlasher(TeensySerial):
	def __init__(self, port):
		TeensySerial.__init__(self, port)

	def ping(self):
		self.write(CMD_PING1)
		self.write(CMD_PING2)
		val = self.readbyte()
		if val != RES_PONG1:
			raise NORError("Ping failed (expected 42, got %02x)"%val)
		val = self.readbyte()
		if val != RES_PONG2:
			raise NORError("Ping failed (expected bd, got %02x)"%val)

	def state(self):
		self.write(CMD_READSTATE)
		return self.readbyte()

	def printstate(self):
		state = self.state()
		print "{0:15} {1}".format("STATUS_DRIVE:", bool(state & STATUS_DRIVE))
		print "{0:15} {1}".format("STATUS_TRIST_N:", bool(state & STATUS_TRIST_N))
		print "{0:15} {1}".format("STATUS_RESET_N:", bool(state & STATUS_RESET_N))
		print "{0:15} {1}".format("STATUS_READY:", bool(state & STATUS_READY))
		print "{0:15} {1}".format("STATUS_CE_N:", bool(state & STATUS_CE_N))
		print "{0:15} {1}".format("STATUS_WE_N:", bool(state & STATUS_WE_N))
		print "{0:15} {1}".format("STATUS_OE_N:", bool(state & STATUS_OE_N))

	def _s_drive(self, v):
		self.write(CMD_DRIVE | bool(v))
	def _g_drive(self):
		return bool(self.state() & STATUS_DRIVE)
	drive = property(_g_drive, _s_drive)

	def _s_trist(self, v):
		self.write(CMD_TRISTATE | bool(v))
	def _g_trist(self):
		return not (self.state() & STATUS_TRIST_N)
	trist = property(_g_trist, _s_trist)

	def _s_reset(self, v):
		self.write(CMD_RESET | bool(v))
	def _g_reset(self):
		return not (self.state() & STATUS_RESET_N)
	reset = property(_g_reset, _s_reset)

	def addr(self, v):
		assert 0 <= v <= 0x7FFFFF
		self.write((CMD_ADDR | (v >> 16), (v >> 8) & 0xff, v & 0xff))

	def wait(self, inc=False):
		self.write(CMD_WAIT | bool(inc))

	def writeword(self, data, inc=False):
		# write [CMD_WRITE | INC, DATA_H, DATA_L]
		self.write((CMD_WRITE | bool(inc), (data>>8) & 0xff, data & 0xff))

	def writeat(self, addr, data, inc=False):
		#write Address
		self.addr(addr)
		#write data
		self.writeword(data, inc)

	def readsector(self, off):
		# set block size
		BLOCK = 0x10000
		# check if offset is a multiple of BLOCK
		assert (off & (BLOCK-1)) == 0
		# write address
		self.addr(off)
		# write CMD_READ AND AUTO-INCREMENT
		self.write(CMD_READ | 0x03)
		# read entire sector
		d = self.read(2*BLOCK)
		# return data readed
		return d

	def read_data_once(self):
		#send CMD_READ_ONCE
		self.write(CMD_READ_ONCE)
		#read D0-D15
		d = self.read(2)
		return d
		
	def wait_for_ry_slow(self):
		time.sleep(0.1)
		while not (self.state() & STATUS_READY):
			time.sleep(0.1)

	def wait_for_ry_fast(self):
		while not (self.state() & STATUS_READY):
			pass

	def erase_samsung(self):
		# erase both chips
		for a22 in [0, 1]:
			print "erasing samsung internal chip %d" % a22
			print "waiting for RY"
			# just to be careful
			self.wait_for_ry_slow()
			print "sending chip erase cmd"
			# chip erase
			self.writeat((a22 << 22) | 0x555, 0xaa)
			self.writeat((a22 << 22) | 0x2aa, 0x55)
			self.writeat((a22 << 22) | 0x555, 0x80)
			self.writeat((a22 << 22) | 0x555, 0xaa)
			self.writeat((a22 << 22) | 0x2aa, 0x55)
			self.writeat((a22 << 22) | 0x555, 0x10)
			print "chip erase cmd done!"
		self.wait_for_ry_slow()
		print "done with samsung flash erase"

	def program_samsung(self, data):
		assert len(data) == 2**24
		# turn data into a word list
		data = struct.unpack(">%dH"%(len(data)/2), data)

		print "programming samsung flash"
		print "waiting for RY"
		self.wait_for_ry_slow()
        
		# program both chips
		for a22 in [0, 1]:
			print "\nprogramming samsung internal chip %d" % a22
			print "unlocking chip"
			# unlock bypass
			self.writeat((a22 << 22) | 0x555, 0xaa)
			self.writeat((a22 << 22) | 0x2aa, 0x55)
			self.writeat((a22 << 22) | 0x555, 0x20)
			print "unlock cmd sent, waiting for RY"
			self.wait_for_ry_slow()
			addr = a22 << 22
			print "setting initial address: 0x%06X" % addr
			# set the addr to the begining of the chip
			self.addr(addr)
			print "starting to program chip %d" % a22
			for addr in range((a22 << 22), (a22 << 22) + 2**22):
                #  + 2**18
				if addr % 512 == 0:
					print "\r%d KB"%(addr/512),
					sys.stdout.flush()
				# unlocked program cmd
				self.writeword(0xa0)
				self.writeword(data[addr], inc=True)
				# included a wait for safety, i bet this will be SLOW
				print "\rWaiting ry_good at 0x%06X(%d KB Done)"%(addr*2,addr/512),
				sys.stdout.flush()
				# self.wait_for_ry_fast()
			print "programming chip %d finished, reseting unlock" % a22
			# reset the unlock
			self.writeword(0x90)
			self.writeword(0x00)
			self.wait_for_ry_fast()
			print "finished chip %d unlock reset" % a22
		print "entire samsung chip was flashed!"

	def erasesector(self, off):
		assert off&0xFFFF == 0
		self.writeat(0x555, 0xaa)
		self.writeat(0x2aa, 0x55)
		self.writeat(0x555, 0x80)
		self.writeat(0x555, 0xaa)
		self.writeat(0x2aa, 0x55)
		self.writeat(off, 0x30)
		self.delay(10)
		self.wait()
		self.ping()

	def programline(self, off, data):
		assert off&0x1f == 0
		if isinstance(data, str):
			data = struct.unpack(">%dH"%(len(data)/2), data)
		assert len(data) <= 32
		saddr = off & ~0x1f
		self.writeat(0x555, 0xaa)
		self.writeat(0x2aa, 0x55)
		self.writeat(saddr, 0x25)
		self.writeat(saddr, len(data)-1)
		self.addr(off)
		for d in data:
			self.writeword(d, True)
		self.writeat(saddr, 0x29)
		self.wait()

	def writesector(self, addr, data):
		assert len(data) == 0x20000
		assert (addr & 0xffff) == 0
		odata = self.readsector(addr)
		if odata == data:
			return
		self.erasesector(addr)
		for off in range(0,0x20000,0x40):
			d = data[off:off+0x40]
			if d != "\xff"*0x40:
				self.programline(addr+(off/2), d)
		rdata = self.readsector(addr)
		if rdata != data:
			raise NORError("Verification failed")

	def rmwsector(self, addr, data):
		offset = addr & 0x1ffff
		endaddr = offset + len(data)
		assert endaddr <= 0x20000
		secaddr = (addr & ~0x1ffff)/2
		odata = self.readsector(secaddr)
		wdata = odata[:offset] + data + odata[endaddr:]
		if odata != wdata:
			self.writesector(secaddr, wdata)

	def writerange(self, addr, data):
		if len(data) == 0:
			return
		first_sector = addr & ~0x1ffff
		last_sector = (addr + len(data) - 1) & ~0x1ffff

		offset = addr & 0x1ffff

		sec_count = (last_sector + 0x20000 - first_sector)/0x20000
		done = 0
		if offset != 0:
			print "\rSector %06x (%d/%d) [F]..."%(first_sector, done+1, sec_count),
			sys.stdout.flush()
			self.rmwsector(addr, data[:0x20000-offset])
			data = data[0x20000-offset:]
			done += 1
			addr += 0x20000-offset

		while len(data) >= 0x20000:
			print "\rSector %06x (%d/%d) [M]..."%(addr, done+1, sec_count),
			sys.stdout.flush()
			self.writesector(addr/2, data[:0x20000])
			done += 1
			addr += 0x20000
			data = data[0x20000:]

		if len(data) != 0:
			print "\rSector %06x (%d/%d) [L]..."%(addr, done+1, sec_count),
			sys.stdout.flush()
			self.rmwsector(addr, data)
			done += 1

		assert done == sec_count
		print

	def delay(self, v):
		while v > 0x41:
			self.write(0x7f)
			v -= 0x41
		if v <= 0:
			return
		elif v == 1:
			self.write(0x00)
		else:
			self.write(0x40 | (v-2))

	def udelay(self, v):
		self.delay(v * 60)

	def checkPPB(self):
		for a22 in range (0 ,1 << 8):
			# PPB Status
			self.writeat(0x555, 0xAA)
			self.writeat(0x2AA, 0x55)
			self.writeat(0x555, 0x90)
			# put bank number in address lines
			self.addr((a22 << 12) | 0x02)
			self.wait_for_ry_fast()

			res = self.read_data_once()
			resH = ord(res[0])
			resL = ord(res[1])
			if (resL & 1) == 0:
				Result = "OFF"
			else:
				Result = "ON"
			print "Bank %03d(%06X) PPB State is %s" % (a22,(a22 << 12),Result)

	def checkDYB(self):
		for a22 in range (0 ,1 << 8):
			# DYB Status
			self.writeat(0x555, 0xAA)
			self.writeat(0x2AA, 0x55)
			self.writeat(0x555, 0x58)
			# put bank number in address lines
			self.addr((a22 << 12) | 0x02)
			self.wait_for_ry_fast()

			res = self.read_data_once()
			resH = ord(res[0])
			resL = ord(res[1])
			if (resL & 1) == 0:
				ResultDYB = "OFF"
			else:
				ResultDYB = "ON"
			if (resL & 2) == 0:
				ResultPPBLB = "OFF"
			else:
				ResultPPBLB = "ON"
			print "Bank %03d(%06X) DYB State is %s, PPBLB is %s" % (a22,(a22 << 12),ResultDYB,ResultPPBLB)

	def writeAlt_old(self):
		# Erase
		# self.erase_samsung()
		# Program
		actualTime = time.time()
		bAddress = 0
		for address in range (0,0x800000):
			self.writeat(0x555,0xAA)
			self.writeat(0x2AA,0x55)
			self.writeat(0x555,0xA0)
			self.writeat(address,data[address])
			self.wait_for_ry_fast()
			if ((time.time() - actualTime >= 1)):
				if address != 0:
					# self.flush()
					actualTime = time.time()
					print "\rWriting...addr:0x%06X - %2.2f%% done(%d Bps)   " % (address, float(address) * 100 / 0x800000,(address - bAddress) * 2),
					sys.stdout.flush()
					bAddress = address
		print "Writing Done!"

	def writeAlt(self, addr):
		# Erase
		# self.erase_samsung()

		# Program
		actualTime = time.time()
		bAddress = addr
		bulkSize = 0x1000
		for address in range (addr, 0x1000000, bulkSize):
			d = data[address:address+bulkSize]

			# self.addr(address)
			# send CMD
			self.write(CMD_WRITE_BULK)
			# send data chunk
			self.write(d)

			# wait for write successful(0x01)
			res = self.readbyte()
			# print "res: %c"%res
			# wait for 'K' char, else error('T' = timeout error when writing, 'R' = Teensy receive buffer timeout)
			assert res == 75

			if ((time.time() - actualTime >= 1)):
				if address != 0:
					actualTime = time.time()
					print "\rWriting...addr:0x%06X - %2.2f%% done(%d KBps)   " % (address / 2, float(address) * 100 / 0x1000000,(address - bAddress) / 512),
					sys.stdout.flush()
					bAddress = address
		print "Writing Done!"

if __name__ == "__main__":
	if len(sys.argv) == 1:
		print "NORway.py v0.3 - Teensy++ 2.0 NOR flasher for PS3 (judges@eEcho.com)"
		print "(Orignal noralizer.py by Hector Martin \"marcan\" <hector@marcansoft.com>)"
		print "(samsung modifications by foobar)"
		print
		print "Usage:"
		print "%s serialport [command] [filename] [address]"%sys.argv[0]
		print
		print "  serialport  Name of serial port to open (eg. COM1, COM2, /dev/ttyACM0, etc)"
		print "  command     dump       Reads entire NOR to [filename]"
		print "              release    Releases NOR interface, so the PS3 can boot"
		print "              erase_samsung   erase the whole samsung chip"
		print "              program_samsung erase,write,verify [filename] to the samsung chip"
		print "  filename    Filename for [dump|program_samsung]"
		print
		print "Examples:"
		print "  %s COM1"%sys.argv[0]
		print "  %s COM1 dump d:\myflash.bin"%sys.argv[0]
		print "  %s COM1 erase_samsung"%sys.argv[0]
		print "  %s COM1 program_samsung d:\myflash.bin"%sys.argv[0]
		print "  %s COM1 release"%sys.argv[0]
		sys.exit(0)
	
	n = NORFlasher(sys.argv[1])
	print "Pinging..."
	n.ping()

	n.drive = 0
	n.reset = 0
	print "Set SB to tristate"
	print
	n.trist = 1
	n.printstate()
	n.drive = 1
	print
	print "Resetting NOR..."
	n.reset = 1
	n.udelay(40)
	n.reset = 0
	n.udelay(40)
	n.ping()
	print "Ready."

	tStart = time.time()
	if len(sys.argv) == 4 and sys.argv[2] == "dump":
		BLOCK = 0x10000
		print
		print "Dumping NOR..."
		fo = open(sys.argv[3],"wb")
		for offset in range(0, 0x800000, BLOCK):
			fo.write(n.readsector(offset))
			print "\r%d KB"%((offset+BLOCK)/512),
			sys.stdout.flush()
		print
		print "Done. [%s]"%(datetime.timedelta(seconds=time.time() - tStart))
	elif len(sys.argv) == 3 and sys.argv[2] == "erase_samsung":
		print
		print "Erasing entire samsung chip"
		sys.stdout.flush()
		n.erase_samsung()
		print "Done. [%s]"%(datetime.timedelta(seconds=time.time() - tStart))
	elif len(sys.argv) == 4 and sys.argv[2] == "program_samsung":
		print
		data = open(sys.argv[3],"rb").read()
		print "erasing entire samsung flash"
		n.erase_samsung()
		print "programming entire samsung flash"
		n.program_samsung(data)
		print "verifying entire samsung flash"
		dv = ""
		BLOCK = 0x10000
		for offset in range(0, 0x800000, BLOCK):
			dv += n.readsector(offset)
			print "\r%d KB"%((offset+BLOCK)/512),
			sys.stdout.flush()
		print

		if d != dv:
			print
			print "Verification failed!"

		print
		print "Done. [%s]"%(datetime.timedelta(seconds=time.time() - tStart))
	elif len(sys.argv) == 3 and sys.argv[2] == "release":
		print
		n.drive = 0
		n.trist = 0
		print "NOR Released"
	elif len(sys.argv) == 4 and sys.argv[2] == "writeAlt_old":
		data = open(sys.argv[3],"rb").read()
		assert len(data) == 2**24
		# turn data into a word list
		data = struct.unpack(">%dH"%(len(data)/2), data)
		n.writeAlt_old()
	elif len(sys.argv) in (4,5) and sys.argv[2] == "writeAlt":
		data = open(sys.argv[3],"rb").read()
		assert len(data) == 2**24
		# turn data into a word list
		# data = struct.unpack(">%dH"%(len(data)/2), data)
		addr = 0
		if len(sys.argv) == 5:
			addr = int(sys.argv[4],16)	
		n.writeAlt(addr)
	elif len(sys.argv) == 4 and sys.argv[2] == "dumpAlt":
		print
		print "Dumping Alternative Method NOR..."
		fo = open(sys.argv[3],"wb")
		for offset in range(0, 0x800000):
			n.addr(offset)
			fo.write(n.read_data_once())
			print "\r%d KB"%(offset/512),
			sys.stdout.flush()
		print
		print "Done. [%s]"%(datetime.timedelta(seconds=time.time() - tStart))
	elif len(sys.argv) == 3 and sys.argv[2] == "checkPPB":
		print
		n.checkPPB()
	elif len(sys.argv) == 3 and sys.argv[2] == "checkDYB":
		print
		n.checkDYB()
	n.ping()
